/*
 * File:   main.c
 * sciLogger サブCPU
 * PIC24Fj64GA004
 * 8MHz
 *
 * Created on February 9, 2010, 10:53 AM
 */

#include <stdio.h>
#include <stdlib.h>
#include <p24FJ64GA004.h>
#include <libpic30.h>
//#include <uart.h>
#include <string.h>
#include "myuart.h"
#include "myspi.h"
#include "debug_print.h"
#include "ublox.h"
#include "myad.h"
#include "delay.h"
#include "ring.h"
#include "byte2var.h"
#include "myint.h"
#include "mysts.h"

_CONFIG1( JTAGEN_OFF & GCP_OFF & GWRP_OFF & COE_OFF & FWDTEN_OFF & ICS_PGx2 )
//_CONFIG2( FCKSM_CSDCMD & OSCIOFNC_ON & POSCMOD_OFF & FNOSC_FRC)
#pragma config FCKSM = CSDCMD, OSCIOFNC = ON, POSCMOD = OFF, FNOSC = FRC, FPBDIV = DIV_8

//#define CPU_CLOCK        10000000                         // クロック[ Hz ]
//#define CPU_PLL          8                                // PLL
//#define CLOCK_FREQUENCY  ( ( CPU_CLOCK * CPU_PLL ) / 4 )  // 動作周波数[ Hz ]
//#define FOSC 4000000L
// UART2 Debug Port
#define BAUDRATE2         9600L
// UART1 GPS
#define BAUDRATE1         9600L

/*** ADバッファ *************************************************
 * 50Hzデータ
 * 1秒平均データ
 * タイムスタンプ
 */
#define ADBUF_BYTE  3   // AD 1CH 1sample dataのbyte数
#define ADBUF_LEN   sizeof(ADbufType)   // 1dataの長さ byte $含まず
#define RESDATA_LEN 17  // コマンドに対する返答データ　最大長さ

// SPI送信にそのまま使われるので注意！
typedef struct {
    unsigned char   rescode;    // コマンドに対する返答 ACK/NAK/NUL
    unsigned char   resdata[RESDATA_LEN];   // コマンドに対する返答データ
    UbloxNavTimeUtc t;  // タイムスタンプ
    long data1sec[AD_CHNUM];  // 1sec平均値
    unsigned char data[AD_CHNUM][ADBUF_BYTE*AD_SAMPLE];  // AD_SAMPLE[Hz]のデータ
    int cheksum;
} ADbufType;
//} __attribute__((__packed__)) ADbufType;
static ADbufType   adbuf[RING_NUM];

// SPI外部送信用バッファ AD_SAMPLE[Hz]
//static char adbuf[RING_NUM][AD_CHNUM][ADBUF_BYTE*AD_SAMPLE];
// 1sec平均値
//static long adbuf_1sec[RING_NUM][AD_CHNUM];
// timestamp
//static UbloxNavTimeUtc adbuf_t[RING_NUM];

#define adbuf_get(buf_no)   (&adbuf[buf_no])
#define adbuf_data_get(buf_no)   (adbuf[buf_no].data)
#define adbuf_1sec_get(buf_no)   (adbuf[buf_no].data1sec)
#define adbuf_time_get(buf_no)   (&(adbuf[buf_no].t))

inline void my_memcpy(void *dest0, void *src0, int len)
{
    int i;
    char *dest, *src;

    dest = (char*)dest0;
    src = (char*)src0;
    for(i = 0; i < len; i++) {
        *dest++ = *src++;
    }
}
inline void my_memset(void *dest0, unsigned char data, int len)
{
    int i;
    char *dest;

    dest = (char*)dest0;
    for(i = 0; i < len; i++) {
        *dest++ = data;
    }
}
inline void adbuf_write(unsigned char buf_no, unsigned char ch, int cnt, long *data)
{
    unsigned char    *ptr = (unsigned char*)data;
    unsigned char   i;

    for(i = 0; i < ADBUF_BYTE; i++) {
        adbuf[buf_no].data[ch][ADBUF_BYTE*cnt + i] = *(ptr++);
    }
}
// 1
inline void adbuf_1sec_write(char buf_no, long *data)
{
    my_memcpy(adbuf_1sec_get(buf_no), data, sizeof(long) * AD_CHNUM);
/*    char ch;
    for(ch = 0; ch < AD_CHNUM; ch++)
        adbuf_1sec[buf_no][ch] = data[ch];
*/
}
inline void adbuf_time_write(char buf_no, UbloxNavTimeUtc *t)
{
//    my_memcpy(&adbuf_t[buf_no], t, sizeof(UbloxNavTimeUtc));
    my_memcpy(adbuf_time_get(buf_no), t, sizeof(UbloxNavTimeUtc));
}

/**** SPI CMD受信バッファ
 */
#define SPI_RX_LEN   20  // Overoから受信するコマンド長　$含む
static char spi_rx_buf[SPI_RX_LEN];
static char spi_rx_cnt; // 受信文字数カウント

#define CMD_STS_NONE    0   // 有効なコマンド無し
#define CMD_STS_RCVD    1   // 有効なコマンド受け取った
static char spi_cmd_sts;    // コマンド受信状態

/**** ublox ************************************************
 */
static UbloxNavTimeUtc gpsNow;
#if 0
void test_ublox_rcv_mon(void)
{
    unsigned char class, id;
    unsigned int    len;
    unsigned char   payload[128];
    char    sz[128];
    UbloxNavTimeUtc *gps, gps1;

    gps = &gps1;
    
    while(1) {
        ublox_rcv_pkt(&class, &id, &len, payload);
        sprintf(sz, "CLASS=%02X ID=%02X LEN=%d\r\n", class, id, len);
        PDEBUG(sz);
        ublox_debug_pkt_out(payload, len);
        ublox_decode_nav_timeutc(payload, gps);
        sprintf(sz, "%04u/%02d/%02d %02d:%02d:%02d.%09ld %6lums %6luns %02X\r\n",
            gps->year, gps->month, gps->day, gps->hour, gps->min, gps->sec, gps->nano,
            gps->tow, gps->tacc, gps->valid
        );
        PDEBUG(sz);

    }
}
#endif
static void ublox_rcv(void)
{
    unsigned char class, id;
    unsigned int    len;
    unsigned char   payload[128];
//    char    sz[128];
    UbloxNavTimeUtc *gps = &gpsNow;

    ublox_rcv_pkt(&class, &id, &len, payload);
//    sprintf(sz, "CLASS=%02X ID=%02X LEN=%d\r\n", class, id, len);
//    PDEBUG(sz);
//    ublox_debug_pkt_out(payload, len);
    if (class == UBX_CLS_NAV && id == UBX_ID_NAV_TIMEUTC && len == UBX_LEN_NAV_TIMEUTC) {
        ublox_decode_nav_timeutc(payload, gps);
/*        sprintf(sz, "%04u/%02d/%02d %02d:%02d:%02d.%09ld %6lums %6luns %02X\r\n",
            gps->year, gps->month, gps->day, gps->hour, gps->min, gps->sec, gps->nano,
            gps->tow, gps->tacc, gps->valid
        );
        PDEBUG(sz);
*/
    }
}
/**** INT1/SPI1 ***********************************************/
/*
 * INT1 Int及びSPI1 IntでADのデータを取得する
 */
#define ADSTS_IDLE  0
#define ADSTS_H 1
#define ADSTS_M 2
#define ADSTS_L 3


static char adint_ch;   // 取得中のCH
static char adint_cnt = 0;  // 0〜AD_SAMPLE-1
static char adint_sts = ADSTS_IDLE;
static long adint_add[AD_CHNUM];    // 1秒平均用加算

/**** 1PPS ********************************************/
static char pps_intf;

#define pps_intf_clear()  pps_intf = 0
#define pps_intf_set()  pps_intf = 1
#define pps_is_intf_enable()    (pps_intf != 0)

/**** DataOut SPI2 *********************************************/
#define SPI2CSpin  (PORTCbits.RC8)

// SPI2 未送信データバイト数
static int  dataout_cnt;
#define dataout_cnt_set(i)  dataout_cnt = (i)
#define dataout_cnt_get()  dataout_cnt
#define dataout_cnt_minus()  (dataout_cnt--)

// 送信位置ポインタ
static char    *dout_ptr;


/**** DRDY OUT ****************************************/
#define DRDY_OUT    PORTCbits.RC9

/**** TEST *******************************************/
/*
 * SPI FLAG test
 */
void test_spi_flag(void)
{
    char    sz[128];
    int in;
    unsigned char txc = 0;
    unsigned char rxc;

//    spi2_int_disable();
//    cnint_int_disable();
    while(1) {
        delay_ms(1000);
        sprintf(sz, "SPI2STAT=%04X "
                "SPIBEC=%d SRMPT=%d SPIROV=%d SRXMPT=%d TBF=%d RBF=%d\r\n"
                , SPI2STAT
                , SPI2STATbits.SPIBEC
                , SPI2STATbits.SRMPT
                , SPI2STATbits.SPIROV
                , SPI2STATbits.SRXMPT
                , SPI2STATbits.SPITBF
                , SPI2STATbits.SPIRBF
                );
        PDEBUG(sz);
        if ((in = uart2_rcvbuf_getc()) >= 0) {
            uart2_myputc(in);
            uart2_puts("\r\n");

            switch(in) {
                case 'r':
                    rxc = spi2_getc();
                    sprintf(sz, "RX=%02X\r\n", rxc);
                    uart2_puts(sz);
                    break;
                case 't':
                    spi2_putc(txc++);
                    sprintf(sz, "TX=%02X\r\n", txc);
                    uart2_puts(sz);
                    break;
                case 'v':
                    spi2_rx_overrun_clear();
                    break;
                case 'x':
                    SPI2STATbits.SPIEN = 0;
                    asm("nop");
                    SPI2STATbits.SPIEN = 1;
                    SPI2BUF;
                    break;
                case 'y':
                    SPI2STATbits.SRXMPT = 1;
                    break;
            }
        }
    }
}
/*
 * polling send/rcv
 */
void test_spi_send(void)
{
    unsigned char txc = 1;
    unsigned char rxc;
    char    sz[128];

    spi2_int_disable();
    cnint_int_disable();
//    SPI2BUF = txc++;
//    rxc = SPI2BUF;
    while(1) {
        while(SPI2STATbits.SPITBF == 0) {
            // TX
            SPI2BUF = txc++;
        }
        if (SPI2STATbits.SRXMPT == 0) {
            while(SPI2STATbits.SRXMPT == 0) {
                // RX FIFO読み出し
                rxc = SPI2BUF;
            }
#if 0
            sprintf(sz, "SPI2STAT=%04X "
                    "SPIBEC=%d SRMPT=%d SPIROV=%d SRXMPT=%d TBF=%d RBF=%d\r\n"
                    , SPI2STAT
                    , SPI2STATbits.SPIBEC
                    , SPI2STATbits.SRMPT
                    , SPI2STATbits.SPIROV
                    , SPI2STATbits.SRXMPT
                    , SPI2STATbits.SPITBF
                    , SPI2STATbits.SPIRBF
                    );
            PDEBUG(sz);
#endif
        }
    }

}
// ADレジスタgetして表示
void test_ad_get_config(void)
{
    int cs = AD_CH1_CS;
    int i;
    unsigned char   reg[9];
    char sz[64];

    ad_cs(cs);
        ad_spi_send(ADCMD_SDATAC);
    ad_cs_dis();
    ad_cs(cs);
        ad_spi_rcv(ADCMD_RREG | 0); // Register read command
        ad_spi_rcv(9 - 1);		 // 9byte write
#if 1
        for(i = 0; i < 9; i++) {
            reg[i] = ad_spi_rcv(0);
        }
#endif
    ad_cs_dis();

    for(i = 0; i < 9; i++) {
        sprintf(sz, "%d: %02X\r\n", i, reg[i]);
        PDEBUG(sz);
    }

}

/*
 * main()
 */
int main(int argc, char** argv) {
    unsigned int i;
    // UART2 DebugPort
    const double Baudrate2 = ( double )FCY / ( 16 * BAUDRATE2 ) - 1;
    // ボーレートの小数点以下を四捨五入する
    unsigned int baudrate2 = ( unsigned int )( Baudrate2 + 0.5 );
    // UART1 GPS
    const double Baudrate1 = ( double )FCY / ( 16 * BAUDRATE1 ) - 1;
    // ボーレートの小数点以下を四捨五入する
    unsigned int baudrate1 = ( unsigned int )( Baudrate1 + 0.5 );
    char    sz[128];
    UbloxNavTimeUtc *gps = &gpsNow;

    /**** 割り込み禁止 ****/
    SET_CPU_IPL( 7 );
    __builtin_write_OSCCONL(OSCCON & 0xbf); //clear the bit 6 of OSCCONL to unlock Pin Re-map
    // 割り込みネストenable
    int_nest_enable();

    // UART1ピン割り当て
    RPINR18bits.U1RXR = 6;	// UART1 RX to RP6
    RPOR2bits.RP5R =  3;       	// UART1 TX(3) to RP5
    // UART2ピン割り当て
    RPINR19bits.U2RXR = 21;	// UART2 RX to RP21
    RPOR10bits.RP20R = 5;       // UART2 TX(5) to RP20
    // PIC-RDYピン割り当て
    RPOR12bits.RP25R = 0;       // RC9

    // DataOut SPI2 pin config
    RPINR22bits.SDI2R = 13;     // SPI2 RX to RP13
    RPINR22bits.SCK2R = 23;     // SPI2 CLK to RP23
    RPINR23bits.SS2R = 24;      // SPI2 SS to RP24
    RPOR11bits.RP22R = 10;      // SPI1 SDO2(10) to RP22
    // AD SPI1 pin config
    RPINR20bits.SDI1R = 18;     // SPI1 RX to RP18
    RPOR8bits.RP16R = 8;        // SPI1 SCK1(8) to RP16
    RPOR8bits.RP17R = 7;        // SPI1 SDO1(7) to RP17
    // TRISx 1=input 0=output
    TRISA = 0x0000;             // PortA RA7,8,9 output
    //        5432109876543210
    TRISB = 0b0011000011000000; // PortB RB13(RP13),RB12,RB7,RB6(RP6) input
    TRISC = 0x01A4;             // ポートC RC8(RP24),RC7(RP23),RC5(RP21),RC2(RP18) input
    CLKDIV = 0;                 // CPU Peripheral Clock Ratio=1:1
    AD1PCFG = 0xFFFF;           // AN1-12input pin = Digital mode

    // INT1(RP12 RB12) AD -DRDY0 setup
    RPINR0bits.INT1R = 12;  // RP12(RB12)
    int1_int_disable();
    int1_level_set(4);    // Int priority
    int1_edge_neg(); // Negative Edge
    // INT0(RB7) GPS +PPS setup
    int0_int_disable();
    int0_level_set(4);    // INT Priority
    int0_edge_pos(); // Positive Edge
    // ChangeNotification INT, CN20(RC8), SPI2(Overo) CS0
    cnint_int_disable();
    CNEN2bits.CN20IE = 1;
    CNPU2bits.CN20PUE = 1;  // Pullup ON
    cnint_level_set(4);

    adint_cnt = 0;
    ad_cs_dis();
    sts_set(STS_NOSYNC);
    pps_intf_clear();
    DRDY_OUT = 1;
    spi_cmd_sts = CMD_STS_NONE;
    
    // GPS UART1初期設定　BRGH=Standard mode
    // 9600bps 8bit nonparity 1stop nonflowcnt.
    //                           5432109876543210
    unsigned int U1MODEvalue = 0b1000100000000000;
    unsigned int U1STAvalue =  0b0000010000000000;
//    OpenUART1( U1MODEvalue, U1STAvalue, baudrate1);
    uart1_init( U1MODEvalue, U1STAvalue, baudrate1);
    // DEBUG UART2初期設定　BRGH=Standard mode
    // 9600bps 8bit nonparity 1stop nonflowcnt.
    //                           5432109876543210
    unsigned int U2MODEvalue = 0b1000100000000000;
    unsigned int U2STAvalue =  0b0000010000000000;
    uart2_init( U2MODEvalue, U2STAvalue, baudrate2);


    spi1_init();
    spi2_init();
    spi1_int_disable();
    spi2_int_disable();
    spi2_putc('$'); // 一番最初のデータ送信のためにセットしておく

    PDEBUG("START\r\n");
    sprintf(sz, "sizeof(ADBUF)=%d\r\n", ADBUF_LEN);
    PDEBUG(sz);

    // GPS UART1 受信割り込み許可
    uart1_rx_int_enable();
    uart1_set_int_level(4);  // 割り込みレベル　デフォルト4
    // DebugOut UART2 受信割り込み許可
    uart2_rx_int_enable();
    uart2_set_int_level(4);  // 割り込みレベル　デフォルト4

    //**** 割り込み許可 **************************************************
    // CPU割り込み優先レベル デフォルト0　ペリフェラルはデフォルト4
//    SRbits.IPL = 2;
    SET_CPU_IPL( 2 );

    // UBXプロトコルのみにする
    ublox_rcvbuf_clear();
    ublox_send_cfg_prt(BAUDRATE1);
    ublox_rcv_ack();
    // TimePulse2にAD用クロック出力する
    ublox_rcvbuf_clear();
    ublox_send_cfg_tp5_timepulse2();
    ublox_rcv_ack();
    // TimePulse1をGPS時刻fixしなくても出力する
    // pulse幅 unlocked=500msec locked=100msec
    ublox_rcvbuf_clear();
    ublox_send_cfg_tp5_timepulse1();
    ublox_rcv_ack();
    // NAV-TIMEUTC出力させる
    ublox_send_cfg_msg();
    ublox_rcv_ack();

    // AD初期化
    ad_reset_dis();
    ad_start_dis();
    ad_cs_dis();
    //
    ad_reset(); // -AD RESET
    // WAIT 2^16 AD system clocks
    delay_ms(100);
    //
    ad_cs_init();
    ad_init();
    pga_init();
    
    ring_init();
    // AD SPI1　割り込み設定のみ　許可しない
    // INT1ハンドラ内で有効にされる
    spi1_intf_clear();
    spi1_set_int_level(4);
    // DataOut SPI2　割り込み許可しない CSのCN INTで許可するので
    spi2_intf_clear();
    spi2_set_int_level(4);

    /**** ペリフェラル割り込み許可 INT1/INT0/CN ************************/
    // INT1割り込み許可
    int1_intf_clear();
    int1_int_enable();    // Int enable
    // INT0割り込み許可
    int0_intf_clear();
    int0_int_enable();    // Int enable
    // ChangeNotification INT許可 SPI2(Overo) CS
    cnint_intf_clear();
    cnint_int_enable();
/****!!!!!!!!!!!!!!!!!!!!!!!!*/
//    test_spi_flag();
//test_spi_send();

    while(1) {
        // GPSから受信してデコード
        ublox_rcv();
        if (gpsNow.valid == 0x07 && sts_get() == STS_NOSYNC) {
            // 時刻有効
            sts_set(STS_SYNCNOW0);
        }
        // DebugOutに1secデータ出力
        if (ring_read2_get() != ring_write_get()) {
            // タイムスタンプget
            gps = adbuf_time_get(ring_read2_get());
            sprintf(sz, "%04u/%02d/%02d %02d:%02d:%02d.%09ld,%6lums,%6luns,%02X",
                gps->year, gps->month, gps->day, gps->hour, gps->min, gps->sec, gps->nano,
                gps->tow, gps->tacc, gps->valid
            );
            PDEBUG(sz);

#if 1
            // AD data get
            char ch;
            long    *lp;
            lp = adbuf_1sec_get(ring_read2_get());
            for(ch=0; ch<AD_CHNUM; ch++) {
                sprintf(sz, ",%+7ld", lp[ch]);
                PDEBUG(sz);
            }
            PDEBUG("\r\n");
#endif
            ring_read2_plus();
        }
        // SPIからコマンド受信
        if (spi_cmd_sts == CMD_STS_RCVD) {
            sprintf(sz, "CMD RCV: %02X %02X %02X %02X\r\n", spi_rx_buf[0], spi_rx_buf[1], spi_rx_buf[2], spi_rx_buf[3]);
            PDEBUG(sz);
            spi_cmd_sts = CMD_STS_NONE;
        }

    }
    
    return (EXIT_SUCCESS);
}
/*
 * INT0 GPS 1PPS割り込みハンドラ
 * 
 *
 */
void __attribute__((interrupt, no_auto_psv, shadow)) _INT0Interrupt(void)
{
    // AD STARTピン=H
    ad_start_ena();

    int0_intf_clear();
    pps_intf_set();
    /****AD同期!!!!!!!!!!!!!!!*/
    // AD個数カウンタ=SAMP_FREQ-1
    adint_cnt = AD_SAMPLE - 1;

    // ADバッファにタイムスタンプ付ける
    adbuf_time_write(ring_write_get(), &gpsNow);

    // AD START信号による同期
    if (sts_get() == STS_SYNCNOW0) {
        // AD STARTピン=L
        ad_start_dis();
        sts_set(STS_SYNCNOW1);
    } else if (sts_get() == STS_SYNCNOW1) {
        // AD STARTピン=H　関数先頭で行っているのでここではやらない
        sts_set(STS_SYNCWAIT);
    } else if (sts_get() == STS_SYNCWAIT) {
        sts_set(STS_SYNCED);
    }
}

/*
 * INT1(RP12) AD DRDY0割り込みハンドラ
 * Negativ Edge
 * SPI1受信開始してSPI1割り込み有効にする
 *
 */
void __attribute__((interrupt, no_auto_psv)) _INT1Interrupt(void)
//void _ISR _INT1Interrupt(void)
{
    int1_intf_clear();
    if (!ad_is_drdy0_enable()) {
        return;
    }

    adint_ch = 0;
    adint_sts = ADSTS_H;
    if (adint_cnt == 0) {
        my_memset(adint_add, 0, sizeof(adint_add));
    }

    ad_cs(adint_ch);    // CSx=L
    spi1_rx_overrun_clear();
    // 前に受信したデータをクリア
    spi1_getc();
    // SPI割り込みON
    spi1_intf_clear();
    spi1_int_enable();
    // 送信
    spi1_putc(0);
}
/*
 * AD SPI1 Interrupt Handler
 * ADから受信したデータをADバッファに格納する
 * 6CHぶん受信したらSPI1割り込みをOFF
 * 
 */
//void _ISR _SPI1Interrupt(void)
void __attribute__((interrupt, auto_psv)) _SPI1Interrupt(void)
{
    static unsigned char in[3];
    unsigned char c;
    long    l;

    spi1_intf_clear();
    c = spi1_getc();
    spi1_rx_overrun_clear();
    switch(adint_sts) {
        case ADSTS_H:
            in[2] = c;
            adint_sts = ADSTS_M;
            spi1_putc(0);
            break;
        case ADSTS_M:
            in[1] = c;
            adint_sts = ADSTS_L;
            spi1_putc(0);
            break;
        case ADSTS_L:
            // CSx=H
            ad_cs_dis();
            in[0] = c;
            l = byte3_to_long(in);
            // バッファに書きこみ
            adbuf_write(ring_write_get(), adint_ch, adint_cnt, &l);
            // 1秒平均用加算
            adint_add[adint_ch] += l;
            // CH進める
            adint_ch++;
            if (adint_ch >= AD_CHNUM) {
                // 全CHデータgetした
                adint_cnt++;
                if (adint_cnt >= AD_SAMPLE) {
                    // 1secぶんのデータgetした
                    // DRDY->H DRDY=Lの時でもDRDYのエッジを作るため
                    DRDY_OUT = 1;
                    // 1sec平均する
                    char ch;
                    for(ch = 0; ch < AD_CHNUM; ch++) {
                        adint_add[ch] /= adint_cnt;
                    }
                    // 1secバッファに書きこみ
                    adbuf_1sec_write(ring_write_get(), adint_add);
                    // バッファ書き込み位置＋１
                    ring_write_plus();
                    adint_cnt = 0;
                    // DRDY->L
// TODO: waitカウンタで制限
                    DRDY_OUT = 0;
                }
                spi1_int_disable(); // SPI割り込みOFF
                adint_sts = ADSTS_IDLE;
            } else {
                // 次のCHへ
                ad_cs(adint_ch);    // CSx=L
                // 送信
                spi1_putc(0);
                adint_sts = ADSTS_H;
            }

            break;
        default:
            spi1_int_disable(); // SPI割り込みOFF
            adint_sts = ADSTS_IDLE;
            break;
    }

}

/*
 * DataOut SPI2 Interrupt Handler
 * TX FIFOに1byteの空きが出来ると発生
 */
void __attribute__((interrupt, no_auto_psv)) _SPI2Interrupt(void)
{
    unsigned char   rx;

    spi2_intf_clear();
    // 送信
    while(SPI2_TBF == 0) {
        if (dataout_cnt_get() <= 0) break;
        spi2_putc(*dout_ptr++);
        dataout_cnt_minus();
    }
    // 受信
    while(SPI2_SRXMPT == 0) {
        rx = spi2_getc();
        if (spi_cmd_sts == CMD_STS_NONE && spi_rx_cnt < SPI_RX_LEN) {
            spi_rx_buf[spi_rx_cnt] = rx;
            spi_rx_cnt++;
            if (spi_rx_cnt >= SPI_RX_LEN) {
                if (spi_rx_buf[0] != 0) {
                    spi_cmd_sts = CMD_STS_RCVD;
                }
            }
        }
    }

    // オーバーランエラーならばクリアする
    if (SPI2STATbits.SPIROV == 1) SPI2STATbits.SPIROV = 0;

}
/*
 * Overo SPI2 CS Input Change Notification
 * Interrupt Handler
 *
 */
void __attribute__((interrupt, no_auto_psv)) _CNInterrupt(void)
{
    cnint_intf_clear();
    if (SPI2CSpin == 1) {
        // CS=H
        // SPI2割り込みOFF
        spi2_int_disable();
        // SPI2をリセットしてTX/RX FIFOクリア
        SPI2STATbits.SPIEN = 0;
        asm("nop");
        SPI2STATbits.SPIEN = 1;
        asm("nop");
        SPI2BUF;    // RBF->0
        // $->TX FIFO
        spi2_putc('$');
        // リングバッファ読み込み位置+1
        ring_read1_plus();
    } else {
        // CS=L
        // リングバッファ読み込み位置get
        dout_ptr = (char*)adbuf_get(ring_read1_get());
        // SPI送信の一番最初
        dataout_cnt_set(ADBUF_LEN);
        // SPI2 TX FIFOに送信データ書き込み
        while(SPI2_TBF == 0) {
            if (dataout_cnt_get() <= 0) break;
            spi2_putc(*dout_ptr++);
            dataout_cnt_minus();
        }
        if (spi_cmd_sts == CMD_STS_NONE) {
            // SPI受信文字数カウンタクリア
            spi_rx_cnt = 0;
        }
        // SPI2割り込みON
        spi2_int_enable();
        // DRDY=H
        DRDY_OUT = 1;
    }
}
#if 0
/*
 * CS0〜15順番にenable
 */
void test_ad_cs()
{
    int i;
    
    while(1) {
        delay_ms(1000);
        ad_cs(i);
        i++;
        if (i >= 16) i = 0;
    }
}

/*
 Dataout SPI2 test
 */
void test_spi2()
{
    unsigned int    c = 0x12;
    unsigned int    rx;
    unsigned int    sts = 0;
    int i;
    
    i = SPI2BUF;
    while(1) {
        PORTCbits.RC9 = 1;
        delay_ms(100);
        PORTCbits.RC9 = 0;
        delay_ms(900);

//        while(spi2_tx_fifo_is_full());
//        spi2_putc(c++);
//        while(spi2_rx_fifo_is_empty());
//        rx = spi2_getc();
#if 0
        while(spi2STATbits.SPIRBF == 0);
        sts = spi2STAT;
        if (spi2STATbits.SPITBF == 0) {
            spi2BUF = c++;
        }
//        while(spi2STATbits.SPIRBF == 1) {
            i = spi2BUF;
//        }
        if (i == '$') {
            c = 0;
        }
        if (spi2STATbits.SPIROV == 1) spi2STATbits.SPIROV = 0;
        sprintf(sz, "%d %04X\r\n", i, sts);
        PDEBUG(sz);
#endif
//        while(spi2STATbits.SPITBF == 1);
//        while(spi2STATbits.SPITBF);
/*
         while(!spi2_tx_fifo_is_full())
            spi2_putc(c++);
        while(!spi2_rx_fifo_is_empty()) {
            rx = spi2_getc();
            if (rx == '$') c = 0;
        }
*/
        //        i = ReadUART1();
//        WriteUART2(i);

//        PDEBUG(mes);
//        delay();
    }
}
#endif
/*
void test_ad1()
{
        unsigned char in[3];
        long l[6];
        char cs;
        while(!ad_is_drdy0_enable());
//        PDEBUG("AD\r\n");
        for(cs=0; cs<6; cs++) {
            ad_cs(cs);
                in[2] = ad_spi_rcv();
                in[1] = ad_spi_rcv();
                in[0] = ad_spi_rcv();
            ad_cs_dis();
            l[cs] = byte3_to_long(in);
//            sprintf(sz, "%02X %02X %02X", in[2], in[1], in[0]);
//            PDEBUG(sz);
        }
        for(cs=0; cs<6; cs++) {
            sprintf(sz, "%+7ld,", l[cs]);
            PDEBUG(sz);
        }
        PDEBUG("\r\n");

}
 */
